
#ifndef _ENGINE_H_
#define _ENGINE_H_

#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "typedefs.h"



#define POLY_VERSION	"0.79"


#define TIMER_SPEED		100




typedef struct {
	int prog_idx;	// PROG index
	float c[3];		// Coordinates of the object
	float angle;	// Angle (yaw)
	float vangle;	// V Angle (pitch) or rotation axis (bspmodel)
	int frame;		// Frame (Model, Sprite) or BSPModel (bspmodel)
	int flags;		// Internal rendering flags (passed to client)
} POB;


typedef struct {
	char prog_name[16];
	int  flags;
	int  (*init)(int obj_idx,char *params);
	int  (*event)(int obj_idx,int event,int param1,int param2);
	void (*process)(POB *pob);
	void (*render)(POB *pob);
	void (*free)(void);
} Prog;


typedef struct {		// Used only in server
	POB pob;
	float speed[4];
	float mins[3], maxs[3];	// bounding box for this object
	int next_frame_tick;
	int is_gravity;
	int leaf_idx;
	int tag;			// Same as 'targetname' in Quake
	int team_next;		// Next object in the team
	byte vars[32];		// Later on there'll be dynamic mem alloc
} Object;



#define ACTION_FORWARD			1
#define ACTION_BACKWARD			2
#define ACTION_MOVE_LEFT		4
#define ACTION_MOVE_RIGHT		8
#define ACTION_TURN_LEFT		16
#define ACTION_TURN_RIGHT		32
#define ACTION_JUMP				128
#define ACTION_ATTACK			256

#define PLAYER_FLAG_DEAD			1
#define PLAYER_FLAG_START_ATTACK	2
#define PLAYER_FLAG_EXTCAMERA		4


typedef struct {		// Used only in server
	char name[16];
	int obj_idx;
	int camobj_idx;
	float head_yaw, head_pitch, head_roll;
	float body_yaw, body_roll;
	float cam_height;
	int action;
	int flags;
	int req;
} Player;

/********************************************************************
			ENGINE.C
*********************************************************************/

#define ENGINE_CLIENT		1
#define ENGINE_SERVER		2


typedef struct {
	HINSTANCE hInstance;	// hInstance of the program
	char dataPath[256];	// Path to the data files
	int mode;		// Mode: client and/or server
	int renderWidth;	// Width for renderer (client)
	int renderHeight;	// Height for renderer (client)
	int renderFlags;	// Renderer flags (client)
} EngineImport;

// Initialize the engine
void Engine_Init(EngineImport *eimp);

// Shutdown the engine
void Engine_Shut(void);

// Process the next engine 'frame'
void Engine_NextFrame(void);

// Is server running?
int  Engine_IsServer(void);

// Is client running?
int  Engine_IsClient(void);

// Engine error
void Engine_Error(char *msg);

// Engine fatal error (terminates the program)
void Engine_FatalError(char *msg);


/********************************************************************
			OTHER STUFF
*********************************************************************/

#include "camera.h"
#include "mybsp.h"
#include "model.h"
#include "sprite.h"
#include "part.h"

#include "client.h"
#include "server.h"



extern int TotalFrames;		// Num frames rendered
extern int TotalTicks;		// Time elapsed
extern int FrameTicks;		// Time taken to process last frame
extern int Done;			// True when finished
extern int Paused;			// True when paused

extern int LocalPlayer;		// Local player if serv/cl on the same machine





/********************************************************************
			WORLD.C
*********************************************************************/

char *Parse_GetKeyValue(char *obj_params,char *key);

void  Entity_NewWorld(BSP *bsp);
void  Entity_PlayerStartCoords(BSP *bsp,float *c);


int   InitWorld(char *name);
void  FreeWorld(void);

int   World_AddObject(int prog_idx,char *params);
void  World_RemoveObject(int obj_idx);
void  World_ObjectMoved(int obj_idx);

int   World_AddPlayer(char *name);
void  World_RemovePlayer(int player_idx);

int  ApplySpeed(int obj_idx, float *pos);
void ApplyAcceleration(int obj_idx, int axis, int dir);
void ApplyGravity(int obj_idx);
void ApplyFriction(int obj_idx);
void ApplyJump(int obj_idx);

int  GetPlayerFromObject(int obj_idx);



/********************************************************************
			NETGAME.C
*********************************************************************/


int  InitNetwork(void);
void ShutNetwork(void);
int  Client_Connect(void);
void Client_Disconnect(void);
void Server_Disconnect(void);


/********************************************************************
			PROG.C
*********************************************************************/

// Register all the progs - server & client - called from inside the engine
void  InitProgs(void);

// Free all resources specific to a prog (not object) - server & client
void  FreeProgs(void);

// Allocate a PROG
Prog *Prog_Alloc(void);

// Init an object with a prog - server
int   Prog_Init(int obj_idx,char *params);

// Send an event to an object's prog - server
int   Prog_Event(int obj_idx,int event,int param1,int param2);

// Process an object - client
void  Prog_Process(POB *pob);

// Render an object - client
void  Prog_Render(POB *pob);


#define PF_TYPE_NONE		1
#define PF_TYPE_BSP_MODEL	2
#define PF_TYPE_MODEL		4
#define PF_TYPE_SPRITE		8
#define PF_TYPE_MASK_ALL	15

#define PF_COL_CHECK		32	// check box and pass (don't change)
#define PF_COL_CLIP			64	// check and clip (don't change)


#define PE_UPDATE			1	// called each frame
#define PE_COLLISION		2	// param1 - other_obj
#define PE_HURT				3	// param1 - attacker, param2 - damage
#define PE_ATTACK			4	// attack (player only)
#define PE_DEATH			5	// die (player only)
#define PE_TAG				6	// param1 - tag_idx, param2 - tag_param
#define PE_TEAM				7	// param1 - src_idx, param2 - notify_param
#define PE_OBJ_WATCH		8	// param1 - old_idx, param2 - new_idx


int   Prog_GetProgByName(char *prog_name);
int   Prog_GetFlags(int obj_idx);
int   Prog_GetProgFlags(int prog_idx);

int   Tag_Register(char *tag_str);
void  Tag_Activate(int tag_idx, int tag_param);

void  Team_AddObject(char *team_str, int obj_idx);
void  Team_RemoveObject(int obj_idx);
void  Team_Notify(int src_idx, int notify_param);

void  Watch_Register(int listener, int target);
void  Watch_Notify(int old_idx, int new_idx);


/********************************************************************
			COLDET.C
*********************************************************************/


int  ClipMovePlayer(BSP *bsp,int obj_idx,float *speed);
void ClipMoveProjectile(BSP *bsp,int obj_idx,float *delta,int skip_idx);
void ClipMoveBSPModel(BSP *bsp,int obj_idx,float *delta,int do_push);
int  CheckRay(BSP *bsp,float *start,float *dir,int skip_obj);



/********************************************************************
			ANIM.C
*********************************************************************/

typedef struct {
	int first, last;
} AnimSequence;

// Returns true if wrap
int  Anim_Loop(AnimSequence *anim,int *pframe,int *ptime);
int  Object_Anim_Loop(AnimSequence *anim,Object *obj);
void Anim_Make(AnimSequence *anim,MDL *mdl,char *name);
int  IsAnimSequence(AnimSequence *anim,int frame);



/********************************************************************
			PAL.C
*********************************************************************/

//extern byte RGBPalette[768];
//extern byte *PalTables[32];
//extern int PalNumTables, PalNumShades, PalNumShades2;

void InitPalette(char *);
void FreePalette(void);


/********************************************************************
			PCX.C
*********************************************************************/

byte *PCX_Decode(byte *pcx,int *width,int *height,byte *pal);
byte *TGA_Decode(byte *tga,int *width,int *height);


/********************************************************************
			DATFILE.C
*********************************************************************/

int   DAT_Open(char *fname);
void  DAT_Close(void);
void *DAT_ReadFile(char *name,int *psize);



/********************************************************************
			FLOAT.C
*********************************************************************/

void   FindPlane(float *,float *,float *,float *);
int    FindPlaneType(float *);
void   FVec_Copy(float *,float *);
void   FVec_AddFVec(float *,float *,float *);
void   FVec_SubFVec(float *,float *,float *);
float  FVec_DotProduct(float *,float *);
void   FVec_CrossProduct(float *,float *,float *);
void   FVec_MulScalar(float *,float,float *);
float  FVec_Normalize(float *);
void   FMatrix_Identity(FMatrix);
void   FMatrix_Copy(FMatrix,FMatrix);
void   FMatrix_MakeRotation(FMatrix,float *);
void   FMatrix_PreXRot(FMatrix,float);
void   FMatrix_PreYRot(FMatrix,float);
void   FMatrix_PreZRot(FMatrix,float);
void   FMatrix_MulFVec(FMatrix,float *,float *);
void   FMatrix_InvMulFVec(FMatrix,float *,float *);
void   FMatrix_MulFMatrix(FMatrix,FMatrix,FMatrix);
void   FMatrix_InvMulFMatrix(FMatrix,FMatrix,FMatrix);


#endif
