/********************************************************************
 R_BSP.C - Render BSP (bit flags)
*********************************************************************/

#include "rsoft.h"


extern int do_light;


static VisInfo vis;
static int draw_sky=0;
BSP *cbsp;


// Bit array of faces. If bit is 1 - face is already in VisFaces[]
static byte VisFaceFlag[(MAX_MAP_FACES>>3)+1];

// List of all (possibly) visible faces for Model #0 (WORLD)
word VisFaces[MAX_MAP_FACES];
int NumVisFaces;
int NumVisLeaves;	// Mostly for debug


void DrawModelVisFaces(int model_idx,float angle);
void MakeWorldVisFaces(void);
void DrawWorldVisFaces(void);
void DrawWorldSky(void);
void AddVisFaces(int leaf_idx);
void DrawFace(int idx);

/**********************************************************************
			DRAW TEXTURED BSP
***********************************************************************/



/*
  Render BSP to camera
*/
void Render_BeginBSP(BSP *bsp)
{
	cbsp=bsp;
	draw_sky=0;

	BSP_MakeVis(bsp,ccam->c,&vis,1);
	MakeWorldVisFaces();

	PR_BeginFrame();
	DrawWorldVisFaces();

}

void Render_BSPModel(POB *pob)
{
	FVec_Copy(pob->c,cbsp->Models[pob->frame].origin);
	DrawModelVisFaces(pob->frame,pob->angle);
}

void Render_EndBSP(void)
{
	if (draw_sky) DrawWorldSky();
	PR_EndFrame();
}



/******************************************************************
			3D CULLING
*******************************************************************/

static dplane_t frustum[5];

/*
  Setup frustum planes in world coords
*/
void FrustumSetup(Camera *C)
{
	float v[3], c,s;
	double a;
	int idx;

	idx=0;

// Min_Z plane
	v[0]=0; v[1]=0; v[2]=-1.0;
	FMatrix_MulFVec(C->ToWorld,v,frustum[idx].normal);
	frustum[idx].dist=FVec_DotProduct(frustum[idx].normal,C->c);
	idx++;


// Horizontal FOV
	a=atan2((double)(C->width>>1),C->proj);
	c=cos(a);
	s=sin(a);

// Left plane
	v[0]=-c; v[1]=0; v[2]=-s;
	FMatrix_MulFVec(C->ToWorld,v,frustum[idx].normal);
	frustum[idx].dist=FVec_DotProduct(frustum[idx].normal,C->c);
	idx++;

// Right plane
	v[0]=c; v[1]=0; v[2]=-s;
	FMatrix_MulFVec(C->ToWorld,v,frustum[idx].normal);
	frustum[idx].dist=FVec_DotProduct(frustum[idx].normal,C->c);
	idx++;

// Vertical FOV
	a=atan2((double)(C->height>>1),C->proj*C->proj_ratio);
	c=cos(a);
	s=sin(a);

// Bottom plane
	v[0]=0; v[1]=-c; v[2]=-s;
	FMatrix_MulFVec(C->ToWorld,v,frustum[idx].normal);
	frustum[idx].dist=FVec_DotProduct(frustum[idx].normal,C->c);
	idx++;

// Top plane
	v[0]=0; v[1]=c; v[2]=-s;
	FMatrix_MulFVec(C->ToWorld,v,frustum[idx].normal);
	frustum[idx].dist=FVec_DotProduct(frustum[idx].normal,C->c);
	idx++;

}

/*
  Check for 3d culling of a box (shorts) against frustum
  Returns true if culled
*/
int FrustumCullBox_short(short *mins, short *maxs)
{
	float p[3], *normal, dot;
	int i,j;

	for(i=0;i<5;i++) {
		dot=-frustum[i].dist;
		normal=frustum[i].normal;
		for(j=0;j<3;j++) {
			if (normal[j]>0) p[j]=(float)mins[j];
			else p[j]=(float)maxs[j];
		}
		dot+=FVec_DotProduct(normal,p);
		if (dot>0) return 1;
	}

	return 0;
}

/*
  Check for 3d culling of a box (floats) against frustum
  Returns true if culled
*/
int FrustumCullBox_float(float *mins, float *maxs)
{
	float p[3], *normal, dot;
	int i,j;

	for(i=0;i<5;i++) {
		dot=-frustum[i].dist;
		normal=frustum[i].normal;
		for(j=0;j<3;j++) {
			if (normal[j]>0) p[j]=mins[j];
			else p[j]=maxs[j];
		}
		dot+=FVec_DotProduct(normal,p);
		if (dot>0) return 1;
	}

	return 0;
}


/******************************************************************
			VISIBLE LISTS
*******************************************************************/

/*
  Make list of all visible faces.
  Decompress leaf vislist, and add all _new_ faces of a visible leaf
  to the list of visible faces.
*/
void MakeWorldVisFaces()
{
	dleaf_t *leaf;
	int i;

	NumVisLeaves=0;
	NumVisFaces=0;

	memset(VisFaceFlag,0,(cbsp->NumFaces+7)>>3);

	FrustumSetup(ccam);

	leaf=&cbsp->Leaves[0];
	for(i=0;i<cbsp->NumLeaves;i++,leaf++) {
		if (vis.VisClusters[leaf->cluster>>3]&(1<<(leaf->cluster&7))) {
			if (!FrustumCullBox_short(leaf->mins,leaf->maxs)) {
				AddVisFaces(i);
				NumVisLeaves++;
			}
		}
	}
}


/*
  Add leaf's faces that are not on the list yet.
*/
void AddVisFaces(int leaf_idx)
{
	dleaf_t *leaf;
	int i,face_idx,idx;
	int firstface, lastface;
	BYTE bit;

	firstface=cbsp->Models[0].firstface;
	lastface=firstface+cbsp->Models[0].numfaces;


	leaf=&cbsp->Leaves[leaf_idx];
	for(i=0;i<leaf->numleaffaces;i++) {
		face_idx=leaf->firstleafface+i;
		face_idx=cbsp->FaceList[face_idx];
		if (face_idx<firstface||face_idx>=lastface) continue;
		idx=face_idx>>3;
		bit=1<<(face_idx&7);
		if ((VisFaceFlag[idx]&bit)==0) {
			VisFaceFlag[idx]|=bit;
			VisFaces[NumVisFaces]=face_idx;
			NumVisFaces++;
		}
	}
}



/******************************************************************
		INTERFACE TO POLY RENDERER
*******************************************************************/

/*
  Draw all visible faces (WORLD only).
*/
void DrawWorldVisFaces()
{
	int i;
	for(i=0;i<NumVisFaces;i++)
		DrawFace(VisFaces[i]);
}

/*
  Draw BSP MODEL faces (no VisFaces calcs).
	Cull the model. Draw ALL model faces
*/
void DrawModelVisFaces(int model_idx,float angle)
{
	FMatrix cam_xform, xform;
	dmodel_t *M;
	float mins[3], maxs[3], cam_coord[3];
	int i;

	M=&cbsp->Models[model_idx];

	FVec_AddFVec(M->mins,M->origin,mins);
	FVec_AddFVec(M->maxs,M->origin,maxs);

	if (FrustumCullBox_float(mins,maxs)) return;

	FVec_Copy(ccam->c,cam_coord);
	FMatrix_Copy(ccam->ToWorld,cam_xform);

	FMatrix_Identity(xform);
	FMatrix_PreZRot(xform,angle);
	FMatrix_MulFMatrix(xform,cam_xform,ccam->ToWorld);

	FVec_SubFVec(ccam->c,M->origin,mins);
	FMatrix_MulFVec(xform,mins,ccam->c);
//	FVec_SubFVec(ccam->c,M->origin,ccam->c);

	for(i=M->firstface;i<M->firstface+M->numfaces;i++)
		DrawFace(i);

	FVec_Copy(cam_coord,ccam->c);
	FMatrix_Copy(cam_xform,ccam->ToWorld);
}


/*
  Do last checks of a face visibility, culling,
  Then add edges to PR
*/
void DrawFace(int idx)
{
	dface_t *face;
	CacheFace *info;
	texinfo_t *texinfo;
	dplane_t *plane;
	dedge_t *e;
	float dot;
	int i,edge_idx;

// Get all necessary pointers
	face=&cbsp->Faces[idx];
	info=&cbsp->CacheFaces[idx];
	plane=&cbsp->Planes[face->planenum];

// Do not draw sky polys
	texinfo=&cbsp->TexInfo[face->texinfo];
	if (texinfo->flags&SURF_SKY) {
		draw_sky=1;
		return;
	}

// Does poly face camera?
	dot=FVec_DotProduct(plane->normal,ccam->c)-plane->dist;
	if (!face->side) { if (dot<0.01) return; }
	else { if (dot>-0.01) return; }

// Add all edges of this poly to global edge table
	PR_BeginBSPFace();
	for(i=0;i<face->numedges;i++) {
		edge_idx=face->firstedge+i;
		edge_idx=cbsp->EdgeList[edge_idx];

		if (edge_idx<0) {
			edge_idx=-edge_idx;
			e=&cbsp->Edges[edge_idx];
			PR_AddBSPEdge(edge_idx,e->v[0],e->v[1]);
		}
		else {
			e=&cbsp->Edges[edge_idx];
			PR_AddBSPEdge(edge_idx,e->v[1],e->v[0]);
		}
	}
	PR_EndBSPFace(idx);
}


static float sky_verts[8][3]={
	{-SKYBOX_SIZE,-SKYBOX_SIZE,-SKYBOX_SIZE},
	{+SKYBOX_SIZE,-SKYBOX_SIZE,-SKYBOX_SIZE},
	{+SKYBOX_SIZE,+SKYBOX_SIZE,-SKYBOX_SIZE},
	{-SKYBOX_SIZE,+SKYBOX_SIZE,-SKYBOX_SIZE},

	{-SKYBOX_SIZE,-SKYBOX_SIZE,+SKYBOX_SIZE},
	{+SKYBOX_SIZE,-SKYBOX_SIZE,+SKYBOX_SIZE},
	{+SKYBOX_SIZE,+SKYBOX_SIZE,+SKYBOX_SIZE},
	{-SKYBOX_SIZE,+SKYBOX_SIZE,+SKYBOX_SIZE}
};

static int sky_faces[6][4]={
	{3,2,6,7},
	{1,0,4,5},
	{0,3,7,4},
	{2,1,5,6},
	{7,6,5,4},
	{0,1,2,3}
};

#define UV_STEP		(256.0/20000.0)

static float sky_uvn[6*3][3]={
	{UV_STEP,0,0}, {0,0,UV_STEP}, { 0,-1, 0},
	{-UV_STEP,0,0}, {0,0,UV_STEP},{ 0, 1, 0},
	{0,UV_STEP,0}, {0,0,UV_STEP}, { 1, 0, 0},
	{0,-UV_STEP,0}, {0,0,UV_STEP},{-1, 0, 0},
	{0,-UV_STEP,0}, {-UV_STEP,0,0}, {0, 0, -1},
	{0,-UV_STEP,0}, {-UV_STEP,0,0}, {0, 0, 1},
};

void DrawWorldSky(void) 
{
	int j,i;


// Draw skybox faces
	for(j=0;j<6;j++) {
		PR_BeginSkyFace();

		PR_AddSkyEdge(
			sky_verts[sky_faces[j][3]],
			sky_verts[sky_faces[j][0]]);

		for(i=0;i<3;i++)
			PR_AddSkyEdge(
				sky_verts[sky_faces[j][i]],
				sky_verts[sky_faces[j][i+1]]);

		PR_EndSkyFace(j,sky_uvn[j*3+0],sky_uvn[j*3+1],sky_uvn[j*3+2]);
	}
}
