/********************************************************************
 SURF.C - Surface caching
*********************************************************************/

#include "rsoft.h"


#define SURF_SIZE		8000*1024

#define MAX_DLIGHTS			256
#define MAX_SURFLIGHTS		16

extern int do_light;

typedef struct {
	float c[3];
	float intensity;
	float radius;
} DLight;

static DLight VisDLights[MAX_DLIGHTS];
int NumVisDLights=0;

typedef struct {
	int face_idx, mip_level;
	int width, height;
	int dynamic;
} CellHeader;

typedef struct {
	int x, y;
	int intensity, radius;
} SurfLight;



static SurfLight SurfLights[MAX_SURFLIGHTS];
static int NumSurfLights=0;


static byte DynamicLightMap[300];

static byte *SurfCache;
static int SurfHead, SurfTail, SurfEnd, SurfSize;


void  Render_AddDLight(float *coord,float *rgb,float intensity,float radius)
{
	DLight *l=&VisDLights[NumVisDLights];

	if (NumVisDLights>=MAX_DLIGHTS) return;
	NumVisDLights++;

	FVec_Copy(coord,l->c);
	l->intensity=intensity;
	l->radius=radius;
}


int MakeSurfLights(CacheFace *info, texinfo_t *texinfo, dplane_t *plane)
{
	DLight *pl=VisDLights;
	SurfLight *sl=SurfLights;
	float u, v, n;
	int i;

	NumSurfLights=0;

	if (NumVisDLights<=0) return 0;

	for(i=0;i<NumVisDLights;i++,pl++) {

		n=FVec_DotProduct(pl->c,plane->normal)-plane->dist;
		n=pl->radius-fabs(n);
		if (n<=1.0) continue;

		u=FVec_DotProduct(pl->c,texinfo->vecs[0])+texinfo->vecs[0][3]-info->min_s;
		if (u<(-pl->radius) || u>(((float)info->size_s)+pl->radius)) 
			continue;
		v=FVec_DotProduct(pl->c,texinfo->vecs[1])+texinfo->vecs[1][3]-info->min_t;
		if (v<(-pl->radius) || v>(((float)info->size_t)+pl->radius)) 
			continue;

		sl->x=(int)u; sl->y=(int)v;
		sl->radius=(int)n;
		sl->intensity=(int)(pl->intensity*255);

		NumSurfLights++; sl++;
	}

	return (NumSurfLights>0);
}


int FastDist(int x1, int y1, int x2, int y2)
{
	if ((x2-=x1)<0) x2=-x2;
	if ((y2-=y1)<0) y2=-y2;
	return(x2+y2-( ((x2>y2)?y2:x2)>>1 ));
}

void ApplySurfLights(int light_width, int light_height)
{
	SurfLight *sl;
	BYTE *light;
	int n, intensity, i, j, x, y, a;

	light=DynamicLightMap;

	for(j=0,y=0;j<light_height;j++,y+=16) {
		for(i=0,x=0;i<light_width;i++,x+=16,light++) {
			intensity=light[0];
			sl=SurfLights;
			for(n=0;n<NumSurfLights;n++,sl++) {
				a=FastDist(x,y,sl->x,sl->y);
				if (a>=sl->radius) continue;
				a=(sl->radius-a)*sl->intensity;
				intensity+=(a>>6);
				if (intensity>=255) {
					intensity=255;
					break;
				}
			}
			light[0]=intensity;
		}
	}
}



int GetSurf(int face_idx, int mip_level, byte **ptr, int *w, int *h)
{
	dface_t *face;
	texinfo_t *texinfo;
	miptex_t *miptex;
	CacheFace *info;
	CellHeader *head;
	BYTE *src_line, *src, *dest, *cache, *pal, *light, *light_line;
	int dest_x, dest_y, src_y, src_x, count_x, light_x;
	int lmap_width, lmap_height, s,ds,weight,u0,v0;
	int width, height, src_width, src_height, size;
	int light_step, light_step2, amb_light, i, j;
	int dynamic, allocate;

// Sky surface
	if (face_idx<0) {
		i=-(face_idx+1);
		*ptr=cbsp->Sky.raw[i];
		*w=cbsp->Sky.width;
		*h=cbsp->Sky.height;
		return 1;
	}

	face=&cbsp->Faces[face_idx];
	info=&cbsp->CacheFaces[face_idx];
	texinfo=&cbsp->TexInfo[face->texinfo];
	miptex=cbsp->Textures[cbsp->TextureList[face->texinfo]].miptex;

// Special surface
	if (texinfo->flags) {
		*w=miptex->width>>mip_level;
		*h=miptex->height>>mip_level;
		*ptr=(byte *)miptex+miptex->offsets[mip_level];
		return 2;
	}

// Check for dynamic lights
	dynamic=MakeSurfLights(info,texinfo,&cbsp->Planes[face->planenum]);

// Try to find if the surf is already in cache
	head=(CellHeader *)(SurfCache+info->cache_offset);
	if (head->face_idx==face_idx&&head->mip_level==mip_level) {
		if (dynamic==FALSE&&head->dynamic==dynamic) {
			*w=head->width;
			*h=head->height;
			*ptr=SurfCache+info->cache_offset+sizeof(CellHeader);
			return 1;
		}
		allocate=FALSE;
	}
	else allocate=TRUE;

// Size of the surface cache entry
	if (info->size_s>256||info->size_t>256) 
		return 0;

	width=info->size_s>>mip_level;
	height=info->size_t>>mip_level;
	*w=width; *h=height;


// Do we need to allocate a new surface
	if (allocate) {

	// Find place for the surface in the cache
		size=width*height+sizeof(CellHeader);
	// Do we need to remove some surfs from cache?
		if (SurfTail+size>SurfHead) {
		// If cache wrap, start from beginning
			if (SurfTail+size>SurfSize) {
				SurfEnd=SurfTail;
				SurfHead=SurfTail=0;
			}
		// Start removing surfaces, until enough space is free
			while(SurfTail+size>SurfHead) {
				if (SurfHead>=SurfEnd) SurfHead=SurfSize;
				else {
					head=(CellHeader *)(SurfCache+SurfHead);
					SurfHead+=head->width*head->height+sizeof(CellHeader);
				}
			}
		}
	// The cache entry starts at SurfTail
		cache=SurfCache+SurfTail;
		info->cache_offset=SurfTail;
		SurfTail+=size;
		if (SurfTail>SurfEnd) SurfEnd=SurfTail;
	}
	else {
		cache=SurfCache+info->cache_offset;
	}

// Init pointers
	head=(CellHeader *)cache; 
	cache+=sizeof(CellHeader);

// Fill out the cell header
	head->face_idx=face_idx;
	head->mip_level=mip_level;
	head->width=width;
	head->height=height;
	head->dynamic=dynamic;

// Start offset of the texture

	u0=info->min_s;
	v0=info->min_t;

	u0%=miptex->width;
	v0%=miptex->height;

	if (u0<0) u0+=miptex->width;
	if (v0<0) v0+=miptex->height;

	u0>>=mip_level;
	v0>>=mip_level;


	src_width=miptex->width>>mip_level;
	src_height=miptex->height>>mip_level;
	src_line=(BYTE *)miptex+miptex->offsets[mip_level];

	light_step=16>>mip_level;
	light_step2=4-mip_level;
	lmap_width=(info->size_s>>4)+1;
	lmap_height=(info->size_t>>4)+1;

	amb_light=255; // face->styles[1];

	dest=cache;


dynamic=1;

// Choose lighting (build lightmap if needed)
	if (dynamic) {
		if (face->lightofs>=0) {
//			memcpy(DynamicLightMap,
//				cbsp->LightMaps+face->lightofs,256);

			src=cbsp->LightMaps+face->lightofs;
			for(j=0;j<lmap_height;j++) {
				for(i=0;i<lmap_width;i++,src+=3) {
					s=src[0];
					if (s<src[1]) s=src[1];
					if (s<src[2]) s=src[2];
					DynamicLightMap[j*lmap_width+i]=s;
				}
			}
		}
		else 
			memset(DynamicLightMap,0,256);
		ApplySurfLights(lmap_width,lmap_height);
		light_line=DynamicLightMap;
	}
	else {
		if (face->lightofs>=0)
			light_line=cbsp->LightMaps+face->lightofs;
		else
			light_line=NULL;
	}

	if (light_line==NULL) {

/////////////////////////////////////
// Constant shading
/////////////////////////////////////

	// Tile the texture into the surface cache
		pal=PalTables[0];

		dest_y=height;
		src_y=src_height-v0;
		src_line+=v0*src_width;

		while(dest_y>0) {

		// Tile horizontal line
			dest_x=width;

			src_x=src_width-u0;
			src=src_line+u0;

			while(dest_x>0) {

				if (src_x<=0) {
					src=src_line;
					src_x=src_width;
				}

				count_x=min(src_x,dest_x);

				dest_x-=count_x;
				src_x-=count_x;

				while(count_x>0) {
					*dest=*src;//pal[*src];
					src++; dest++;
					count_x--;
				}
			}

		// Advance ptr to the src_line
			src_y--; src_line+=src_width;
		// src_line wrap
			if (src_y<=0) {
				src_line-=src_width*src_height;
				src_y=src_height;
			}

			dest_y--;
		}
	}
	else {

/////////////////////////////////////
// Light map shading
/////////////////////////////////////

	// Tile the texture into the surface cache

		pal=PalTables[0];

		dest_y=height;
		src_y=src_height-v0;
		src_line+=v0*src_width;
		weight=light_step;

		while(dest_y>0) {

		// Tile horizontal line
			dest_x=width;
			light=light_line;

			ds=(light[0]*weight+
			    light[lmap_width]*(light_step-weight)
			   )>>light_step2;
			ds=(amb_light-ds)<<(8+PalNumShades2);
			light_x=0;
			s=ds;

			src_x=src_width-u0;
			src=src_line+u0;

			while(dest_x>0) {

				if (light_x<=0) {
					light++; light_x=light_step;
					ds=(light[0]*weight+
					    light[lmap_width]*(light_step-weight)
					   )>>light_step2;
					ds=(amb_light-ds)<<(8+PalNumShades2);
					ds=(ds-s)>>light_step2;
				}
				if (src_x<=0) {
					src=src_line;
					src_x=src_width;
				}

				count_x=min(src_x,dest_x);
				if (count_x>light_x) count_x=light_x;

				dest_x-=count_x;
				src_x-=count_x;
				light_x-=count_x;

				while(count_x>0) {
					*dest=pal[((s>>16)<<8)+(*src)];
					src++; dest++;
					s+=ds;
					count_x--;
				}
			}

		// Advance ptr to the src_line
			src_y--; src_line+=src_width;
		// src_line wrap
			if (src_y<=0) {
				src_line-=src_width*src_height;
				src_y=src_height;
			}

			dest_y--;
			weight--;
			if (weight==0) {
				weight=light_step;
				light_line+=lmap_width;
			}
		}
	}

//	ShowCache(head);

	*ptr=cache;

	return 1;
}
/*
void ShowCache(CellHeader *head)
{
	BYTE *ptr;
	int i, j, w, h;

	ptr=(BYTE *)head;
	ptr+=sizeof(CellHeader);

	w=min(ScrWidth,head->width);
	h=min(ScrHeight,head->height);

	for(j=0;j<h;j++)
		for(i=0;i<w;i++)
			Screen[j*ScrWidth+i]=ptr[j*head->width+i];
	getch();

	while(h>0) {
		h--;
		memset(&Screen[h*ScrWidth],0,w);
	}
}
*/

void InitSurfCache(void)
{
	SurfSize=SURF_SIZE;

	SurfCache=NULL;
	while(!SurfCache&&SurfSize>1024*1024) {
		SurfCache=malloc(SurfSize);
		if (!SurfCache) SurfSize-=1024*1024;
	}

	if (!SurfCache) {
		Engine_FatalError("Could not alloc surf cache.");
	}

	SurfTail=0;
	SurfEnd=0;
	SurfHead=SurfSize;

	memset(SurfCache,0xFF,SurfSize);
}

void ShutSurfCache(void)
{
	if (SurfCache) free(SurfCache);
}
