/********************************************************************
 WORLD.C - World management
*********************************************************************/

#include "engine.h"
#include "progs\progs.h"


// The WORLD

BSP WorldBSP;
char WorldBSPName[64];

// All the objects in the world
Object Objects[MAX_OBJECTS];
int NumObjects=0;

// All the players (clients) in the world
Player Players[MAX_PLAYERS];
int NumPlayers=0;


/*
 Initialize a new world.
 Returns: non-zero on success
*/
int  InitWorld(char *name)
{
// Just in case
	FreeWorld();

// Init particles
	InitParticles();

// Load bsp file for world
	if (!BSP_Load(&WorldBSP,name)) return 0;
	strcpy(WorldBSPName,name);
	return 1;
}

/*
 Free all world resources
*/
void FreeWorld(void)
{
	NumObjects=0;
	BSP_Free(&WorldBSP);
	FreeProgs();
}



int World_AddObject(int prog_idx,char *params)
{
	Object *o=&Objects[NumObjects];
	int i;

	o->pob.prog_idx=prog_idx;
	for(i=0;i<3;i++) o->pob.c[i]=0;
	o->pob.angle=0;
	o->pob.vangle=0;
	o->pob.frame=0;
	o->pob.flags=0;
	for(i=0;i<4;i++) o->speed[i]=0;
	for(i=0;i<3;i++) o->mins[i]=o->maxs[i]=0;

	o->is_gravity=0;
	o->tag=-1;
	o->team_next=-1;

	NumObjects++;

	if (Prog_Init(NumObjects-1,params)) {
		NumObjects--;
		return -1;
	}

	return(NumObjects-1);
}

void World_RemoveObject(int idx)
{
	int i;

	if (idx<0||idx>=NumObjects) return;

// Remove object from its team
	Team_RemoveObject(idx);

// Dec number of objects
	NumObjects--;
	Objects[idx]=Objects[NumObjects];

// Notify the watches
	Watch_Notify(idx,-1);
	Watch_Notify(NumObjects,idx);

// Find players that use pob[NumObjects]
	for(i=0;i<NumPlayers;i++) {
		if (Players[i].obj_idx==NumObjects)
			Players[i].obj_idx=idx;
		if (Players[i].camobj_idx==NumObjects)
			Players[i].camobj_idx=idx;
	}
}


int  World_AddPlayer(char *name)
{
	Player *p;

	p=&Players[NumPlayers];
	NumPlayers++;

// Player has no object and no camera object
	p->obj_idx=-1;
	p->camobj_idx=-1;
// Zero angles and camera height
	p->body_yaw=p->body_roll=0;
	p->head_yaw=p->head_pitch=p->head_roll=0;
	p->cam_height=0;
// Player has no action and is alive
	p->action=0;
	p->flags=0;
// Set player name
	strcpy(p->name,name);
// Set req (server: for update, client: for vis info)
	p->req=0;

printf("Add player %d - %s\n",NumPlayers-1,p->name);

	return(NumPlayers-1);
}

void World_RemovePlayer(int idx)
{
	Player *p;

	p=&Players[idx];

	printf("Remove player %d\n",idx);

// Remove player object (if any)
	if (p->obj_idx>=0) World_RemoveObject(p->obj_idx);
// Dec number of players
	NumPlayers--;
// Copy last player in place of the deleted one
	*p=Players[NumPlayers];
}


int GetPlayerFromObject(int obj_idx)
{
	int i;

	for(i=0;i<NumPlayers;i++) {
		if (Players[i].obj_idx==obj_idx) return i;
	}
	return -1;
}



/*********************************************************************
			PHYSICS
**********************************************************************/

// All speeds, accelerations are per tick

#define MAX_SPEED		2.5
#define C_ACC			0.5
#define C_FRIC			0.2
#define C_GRAVITY		0.04	//0.05
#define JUMP_SPEED		2.0	//4.0
#define TURN_SPEED		(PI/100.0)

/*
 Returns non-zero if coords changed
*/
int ApplySpeed(int obj_idx, float *delta)
{
	Object *o=&Objects[obj_idx];
	FMatrix M;
	float *p=o->speed;
	float ticks=(float)FrameTicks;

	o->pob.angle+=p[3]*ticks;

	if (p[0]!=0||p[1]!=0||p[2]!=0) {
		FMatrix_Identity(M);
		FMatrix_PreZRot(M,o->pob.angle);
		FMatrix_MulFVec(M,p,delta);
		FVec_MulScalar(delta,ticks,delta);
		return 1;
	}
	else {
		delta[0]=delta[1]=delta[2]=0.0;
	}
	return 0;
}

void ApplyGravity(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	float ticks=(float)FrameTicks;
	float d=JUMP_SPEED*C_GRAVITY*ticks*0.5;

	if (o->is_gravity) 
		o->speed[2]-=d;
}

void ApplyJump(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	float d=JUMP_SPEED;

//	if (!o->is_gravity)
		o->speed[2]=d;
}

void ApplyAcceleration(int obj_idx,int axis,int dir)
{
	Object *o=&Objects[obj_idx];
	float ticks=1.0;//(float)FrameTicks;
	float d=MAX_SPEED*C_ACC*ticks;
	float t=TURN_SPEED*C_ACC*ticks;

	if (axis<3)
		o->speed[axis]+=dir<0?-d:d;
	else
		o->speed[axis]=dir<0?-t*2:t*2;
}

void ApplyFriction(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	float ticks=1.0;//(float)FrameTicks;
	float *p=o->speed;
	float m=MAX_SPEED, d=MAX_SPEED*C_FRIC*ticks;
	float t=TURN_SPEED*C_FRIC*ticks;
	int i;

	for(i=0;i<2;i++) {
		if (p[i]<0) {
			p[i]+=d;
			if (p[i]<-m) p[i]=-m;
			else if (p[i]>0) p[i]=0;
		}
		else if (p[i]>0) {
			p[i]-=d;
			if (p[i]>m) p[i]=m;
			else if (p[i]<0) p[i]=0;
		}
	}

	if (p[3]<0) {
		p[3]+=t;
		if (p[3]>0) p[3]=0;
	}
	else if (p[3]>0) {
		p[3]-=t;
		if (p[3]<0) p[3]=0;
	}
}


void World_ObjectMoved(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	float coords[3];
	int idx;

char str[64];

	if (Prog_GetFlags(obj_idx)&PF_TYPE_BSP_MODEL) {
		FVec_AddFVec(WorldBSP.Models[o->pob.frame].maxs,
			WorldBSP.Models[o->pob.frame].mins,coords);
		FVec_MulScalar(coords,0.5,coords);
		FVec_AddFVec(coords,WorldBSP.Models[o->pob.frame].origin /*o->pob.c*/,coords);

//		FVec_Copy(WorldBSP.Models[o->pob.frame].origin /*o->pob.c*/,coords);
		idx=BSP_FindLeaf(&WorldBSP,coords);

//sprintf(str,"bsp %d",idx);
//MessageBox(NULL,str,"info",MB_OK);

	
	}
	else {
		idx=BSP_FindLeaf(&WorldBSP,o->pob.c);
	}

	if (idx>=0) o->leaf_idx=idx;
}


/*********************************************************************
			MISC OBJECT FUNCTIONS
**********************************************************************/

/*
void Object_GetBoundBox(int obj_idx,float *mins,float *maxs)
{
	FMatrix M;
	float omin[3], omax[3], v[3], t[3];
	Object *o=&Objects[obj_idx];
	MDL *mdl;
	float *scale;
	float *origin;
	daliasframe_t *frame;
	int i;

	Prog_GetPointer(&o->pob,&mdl);

	FVec_Copy(o->pob.c,mins);
	FVec_Copy(o->pob.c,maxs);
*/
/*
	scale=(float *)mdl->Header->scale;
	origin=(float *)mdl->Header->scale_origin;

	frame=(daliasframe_t *)(mdl->Frames[o->pob.frame]+4);

	for(i=0;i<3;i++) {
		omin[i]=origin[i]+scale[i]*(float)frame->bboxmin.v[i];
		omax[i]=origin[i]+scale[i]*(float)frame->bboxmax.v[i];
	}

	mins[2]=omin[2]; maxs[2]=omax[2];

	FMatrix_Identity(M);
	FMatrix_PreZRot(M,o->pob.angle+PI/2);

	v[0]=omin[0]; v[1]=omin[1];
	FMatrix_MulFVec(M,v,t);
	mins[0]=t[0]; mins[1]=t[1];
	maxs[0]=t[0]; maxs[1]=t[1];

	v[0]=omin[0]; v[1]=omax[1];
	FMatrix_MulFVec(M,v,t);
	for(i=0;i<2;i++) {
		if (t[i]<mins[i]) mins[i]=t[i];
		else if (t[i]>maxs[i]) maxs[i]=t[i];
	}

	v[0]=omax[0]; v[1]=omin[1];
	FMatrix_MulFVec(M,v,t);
	for(i=0;i<2;i++) {
		if (t[i]<mins[i]) mins[i]=t[i];
		else if (t[i]>maxs[i]) maxs[i]=t[i];
	}

	v[0]=omax[0]; v[1]=omax[1];
	FMatrix_MulFVec(M,v,t);
	for(i=0;i<2;i++) {
		if (t[i]<mins[i]) mins[i]=t[i];
		else if (t[i]>maxs[i]) maxs[i]=t[i];
	}

	FVec_AddFVec(mins,o->pob.c,mins);
	FVec_AddFVec(maxs,o->pob.c,maxs);
*/
//}

